"""
pygame-menu
https://github.com/ppizarror/pygame-menu

WIDGET
This module contains the widgets of pygame-menu.
"""

from pygame_menu.widgets.widget.button import Button
from pygame_menu.widgets.widget.colorinput import ColorInput
from pygame_menu.widgets.widget.dropselect import DropSelect
from pygame_menu.widgets.widget.dropselect_multiple import DropSelectMultiple
from pygame_menu.widgets.widget.frame import Frame
from pygame_menu.widgets.widget.hmargin import HMargin
from pygame_menu.widgets.widget.image import Image
from pygame_menu.widgets.widget.label import Label
from pygame_menu.widgets.widget.menubar import MenuBar
from pygame_menu.widgets.widget.menulink import MenuLink
from pygame_menu.widgets.widget.none import NoneWidget
from pygame_menu.widgets.widget.progressbar import ProgressBar
from pygame_menu.widgets.widget.rangeslider import RangeSlider
from pygame_menu.widgets.widget.scrollbar import ScrollBar
from pygame_menu.widgets.widget.selector import Selector
from pygame_menu.widgets.widget.surface import SurfaceWidget
from pygame_menu.widgets.widget.table import Table
from pygame_menu.widgets.widget.textinput import TextInput
from pygame_menu.widgets.widget.toggleswitch import ToggleSwitch
from pygame_menu.widgets.widget.vfill import VFill
from pygame_menu.widgets.widget.vmargin import VMargin
