"""
pygame-menu
https://github.com/ppizarror/pygame-menu

SELECTION
This module contains the widget highlight effects.
"""

from pygame_menu.widgets.selection.highlight import HighlightSelection
from pygame_menu.widgets.selection.left_arrow import LeftArrowSelection
from pygame_menu.widgets.selection.none import NoneSelection
from pygame_menu.widgets.selection.right_arrow import RightArrowSelection
from pygame_menu.widgets.selection.simple import SimpleSelection
